# -*- coding: utf-8 -*-

# Copyright (c) 2017. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from lifelines import KaplanMeierFitter, CoxPHFitter
from lifelines.statistics import logrank_test
import logging
import matplotlib.colors as colors
from matplotlib import pyplot as plt
import numbers
import numpy as np
import seaborn as sb
import patsy
from .rounding import float_str
from .utils import get_logger

logger = get_logger(__name__, level=logging.INFO)

def _plot_kmf_single(df,
                     condition_col,
                     survival_col,
                     censor_col,
                     threshold,
                     title,
                     xlabel,
                     ylabel,
                     ax,
                     with_condition_color,
                     no_condition_color,
                     with_condition_label,
                     no_condition_label,
                     color_map,
                     label_map,
                     color_palette,
                     ci_show,
                     print_as_title):
    """
    Helper function to produce a single KM survival plot, among observations in df by groups defined by condition_col.

    All inputs are required - this function is intended to be called by `plot_kmf`.
    """
    # make color inputs consistent hex format
    if colors.is_color_like(with_condition_color):
        with_condition_color = colors.to_hex(with_condition_color)
    if colors.is_color_like(no_condition_color):
        no_condition_color = colors.to_hex(no_condition_color)
    ## prepare data to be plotted; producing 3 outputs:
    # - `condition`, series containing category labels to be plotted
    # - `label_map` (mapping condition values to plot labels)
    # - `color_map` (mapping condition values to plotted colors)
    if threshold is not None:
        is_median = threshold == "median"
        if is_median:
            threshold = df[condition_col].median()
        label_suffix = float_str(threshold)
        condition = df[condition_col] > threshold
        default_label_no_condition = "%s ≤ %s" % (condition_col, label_suffix)
        if is_median:
            label_suffix += " (median)"
        default_label_with_condition = "%s > %s" % (condition_col, label_suffix)
        with_condition_label = with_condition_label or default_label_with_condition
        no_condition_label = no_condition_label or default_label_no_condition
        if not label_map:
            label_map = {False: no_condition_label,
                         True: with_condition_label}
        if not color_map:
            color_map = {False: no_condition_color,
                         True: with_condition_color}
    elif df[condition_col].dtype == 'O' or df[condition_col].dtype.name == "category":
        condition = df[condition_col].astype("category")
        if not label_map:
            label_map = dict()
            [label_map.update({condition_value: '{} = {}'.format(condition_col,
                                                        condition_value)})
                     for condition_value in condition.unique()]
        if not color_map:
            rgb_values = sb.color_palette(color_palette, len(label_map.keys()))
            hex_values = [colors.to_hex(col) for col in rgb_values]
            color_map = dict(zip(label_map.keys(), hex_values))
    elif df[condition_col].dtype == 'bool':
        condition = df[condition_col]
        default_label_with_condition = "= {}".format(condition_col)
        default_label_no_condition = "¬ {}".format(condition_col)
        with_condition_label = with_condition_label or default_label_with_condition
        no_condition_label = no_condition_label or default_label_no_condition
        if not label_map:
            label_map = {False: no_condition_label,
                         True: with_condition_label}
        if not color_map:
            color_map = {False: no_condition_color,
                         True: with_condition_color}
    else:
        raise ValueError('Don\'t know how to plot data of type\
                         {}'.format(df[condition_col].dtype))

    # produce kmf plot for each category (group) identified above
    kmf = KaplanMeierFitter()
    grp_desc = list()
    grp_survival_data = dict()
    grp_event_data = dict()
    grp_names = list(condition.unique())
    for grp_name, grp_df in df.groupby(condition):
        grp_survival = grp_df[survival_col]
        grp_event = (grp_df[censor_col].astype(bool))
        grp_label = label_map[grp_name]
        grp_color = color_map[grp_name]
        kmf.fit(grp_survival, grp_event, label=grp_label)
        desc_str = "# {}: {}".format(grp_label, len(grp_survival))
        grp_desc.append(desc_str)
        grp_survival_data[grp_name] = grp_survival
        grp_event_data[grp_name] = grp_event
        if ax:
            ax = kmf.plot(ax=ax, show_censors=True, ci_show=ci_show, color=grp_color)
        else:
            ax = kmf.plot(show_censors=True, ci_show=ci_show, color=grp_color)

    ## format the plot
    # Set the y-axis to range 0 to 1
    ax.set_ylim(0, 1)
    y_tick_vals = ax.get_yticks()
    ax.set_yticklabels(["%d" % int(y_tick_val * 100) for y_tick_val in y_tick_vals])
    # plot title
    if title:
        ax.set_title(title)
    elif print_as_title:
        ax.set_title(' | '.join(grp_desc))
    else:
        [print(desc) for desc in grp_desc]
    # axis labels
    if xlabel:
        ax.set_xlabel(xlabel)
    if ylabel:
        ax.set_ylabel(ylabel)
    
    ## summarize analytical version of results
    ## again using same groups as are plotted
    if len(grp_names) == 2:
        # use log-rank test for 2 groups
        results = logrank_test(grp_survival_data[grp_names[0]],
                               grp_survival_data[grp_names[1]],
                               event_observed_A=grp_event_data[grp_names[0]],
                               event_observed_B=grp_event_data[grp_names[1]])
    elif len(grp_names) == 1:
        # no analytical result for 1 or 0 groups
        results = NullSurvivalResults()
    else:
        # cox PH fitter for >2 groups
        cf = CoxPHFitter()
        cox_df = patsy.dmatrix('+'.join([condition_col, survival_col,
                                         censor_col]),
                               df, return_type='dataframe')
        del cox_df['Intercept']
        results = cf.fit(cox_df, survival_col, event_col=censor_col)
        results.print_summary()
    # add metadata to results object so caller can print them
    results.survival_data_series = grp_survival_data
    results.event_data_series = grp_event_data
    results.desc = grp_desc
    return results


def plot_kmf(df,
             condition_col,
             censor_col,
             survival_col,
             strata_col=None,
             threshold=None,
             title=None,
             xlabel=None,
             ylabel=None,
             ax=None,
             with_condition_color="#B38600",
             no_condition_color="#A941AC",
             with_condition_label=None,
             no_condition_label=None,
             color_map=None,
             label_map=None,
             color_palette="Set1",
             ci_show=False,
             print_as_title=False):
    """
    Plot survival curves by splitting the dataset into two groups based on
    condition_col. Report results for a log-rank test (if two groups are plotted)
    or CoxPH survival analysis (if >2 groups) for association with survival.

    Regarding definition of groups:
        If condition_col is numeric, values are split into 2 groups.
             - if threshold is defined, the groups are split on being > or < condition_col
             - if threshold == 'median', the threshold is set to the median of condition_col
        If condition_col is categorical or string, results are plotted for each unique value in the dataset.
        If condition_col is None, results are plotted for all observations

    Currently, if `strata_col` is given, the results are repeated among each stratum of the df.
    A truly "stratified" analysis is not yet supported by may be soon.

    Parameters
    ----------
        df: dataframe
        condition_col: string, column which contains the condition to split on
        survival_col: string, column which contains the survival time
        censor_col: string,
        strata_col: optional string, denoting column containing data to
                    stratify by (default: None)
        threshold: int or string, if int, condition_col is thresholded at int,
                                  if 'median', condition_col thresholded
                                  at its median
                                  if 'median-per-strata', & if stratified analysis
                                  then condition_col thresholded by strata
        title: Title for the plot, default None
        ax: an existing matplotlib ax, optional, default None
             note: not currently supported when `strata_col` is not None
        with_condition_color: str, hex code color for the with-condition curve
        no_condition_color: str, hex code color for the no-condition curve
        with_condition_label: str, optional, label for True condition case
        no_condition_label: str, optional, label for False condition case
        color_map: dict, optional, mapping of hex-values to condition text
          in the form of {value_name: color_hex_code}.
          defaults to `sb.color_palette` using `default_color_palette` name,
          or *_condition_color options in case of boolean operators.
        label_map: dict, optional, mapping of labels to condition text.
          defaults to "condition_name = condition_value", or *_condition_label
          options in case of boolean operators.
        color_palette: str, optional, name of sb.color_palette to use
          if color_map not provided.
        print_as_title: bool, optional, whether or not to print text
          within the plot's title vs. stdout, default False
    """
    
    # set reasonable default threshold value depending on type of condition_col
    if threshold is None:
        if df[condition_col].dtype != "bool" and \
            np.issubdtype(df[condition_col].dtype, np.number):
                threshold = "median"
    # check inputs for threshold for validity
    elif isinstance(threshold, numbers.Number):
        logger.debug("threshold value is numeric")
    elif threshold not in ("median", "median-per-strata"):
        raise ValueError("invalid input for threshold. Must be numeric, None, 'median', or 'median-per-strata'.")
    elif threshold == "median-per-strata" and strata_col is None:
        raise ValueError("threshold given was 'median-per-strata' and yet `strata_col` was None. Did you mean 'median'?")

    # construct kwarg dict to pass to _plot_kmf_single.
    # start with args that do not vary according to strata_col
    arglist = dict(
            condition_col=condition_col,
            survival_col=survival_col,
            censor_col=censor_col,
            threshold=threshold,
            with_condition_color=with_condition_color,
            no_condition_color=no_condition_color,
            with_condition_label=with_condition_label,
            no_condition_label=no_condition_label,
            color_map=color_map,
            label_map=label_map,
            xlabel=xlabel,
            ylabel=ylabel,
            ci_show=ci_show,
            color_palette=color_palette,
            print_as_title=print_as_title)

    # if strata_col is None, pass all parameters to _plot_kmf_single
    if strata_col is None:
        arglist.update(dict(
            df=df,
            title=title,
            ax=ax))
        return _plot_kmf_single(**arglist)
    else:
        # prepare for stratified analysis
        if threshold == "median":
            # by default, "median" threshold should be intra-strata median
            arglist["threshold"] = df[condition_col].dropna().median()
        elif threshold == "median-per-strata":
            arglist["threshold"] = "median"
        # create axis / subplots for stratified results
        if ax is not None:
            raise ValueError("ax not supported with stratified analysis.")
        n_strata = len(df[strata_col].unique())
        f, ax = plt.subplots(n_strata, sharex=True)
        # create results dict to hold per-strata results
        results = dict()
        # call _plot_kmf_single for each of the strata
        for i, (strat_name, strat_df) in enumerate(df.groupby(strata_col)):
            if n_strata == 1:
                arglist["ax"] = ax
            else:
                arglist["ax"] = ax[i]
            subtitle = "{}: {}".format(strata_col, strat_name)
            arglist["title"] = subtitle
            arglist["df"] = strat_df
            results[subtitle] = plot_kmf(**arglist)
            [print(desc) for desc in results[subtitle].desc]
        if title:
            f.suptitle(title)
        return results


class NullSurvivalResults(object):
    def __repr__(self):
        return "No model fit."


def logrank(df,
            condition_col,
            censor_col,
            survival_col,
            threshold=None):
    if threshold is not None:
        if threshold == "median":
            threshold = df[condition_col].median()
        condition = df[condition_col] > threshold
    else:
        condition = df[condition_col]
    df_with_condition = df[condition]
    df_no_condition = df[~condition]
    survival_no_condition = df_no_condition[survival_col]
    survival_with_condition = df_with_condition[survival_col]
    event_no_condition = (df_no_condition[censor_col].astype(bool))
    event_with_condition = (df_with_condition[censor_col].astype(bool))
    return logrank_test(survival_no_condition,
                        survival_with_condition,
                        event_observed_A=event_no_condition,
                        event_observed_B=event_with_condition)



# Copyright (c) 2017. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .variant_stats import variant_stats_from_variant
from .utils import get_logger

from varcode import Variant
from varcode.common import memoize
import pandas as pd
from os import path

logger = get_logger(__name__)

def no_filter(filterable_variant):
    return True

def variant_qc_filter(filterable_variant,
                      min_tumor_depth,
                      min_normal_depth,
                      min_tumor_vaf,
                      max_normal_vaf,
                      min_tumor_alt_depth):
    logger.debug('Applying variant_qc_filter with params: min_tumor_depth={}, min_normal_depth={}, min_tumor_vaf={}, max_normal_vaf={}, min_tumor_alt_depth={}'.format(min_tumor_depth, min_normal_depth, min_tumor_vaf, max_normal_vaf, min_tumor_alt_depth))

    somatic_stats = variant_stats_from_variant(filterable_variant.variant,
                                               filterable_variant.variant_metadata)

    # Filter variant with depth < depth
    if (somatic_stats.tumor_stats.depth < min_tumor_depth or
        somatic_stats.normal_stats.depth < min_normal_depth):
        return False

    # Filter based on normal evidence
    if somatic_stats.normal_stats.variant_allele_frequency > max_normal_vaf:
        return False

    if somatic_stats.tumor_stats.variant_allele_frequency < min_tumor_vaf:
        return False

    if somatic_stats.tumor_stats.alt_depth < min_tumor_alt_depth:
        return False

    return True

@memoize
def expressed_variant_set(cohort, patient, variant_collection):
    # Warning: we previously had an issue where we used the same
    # memoized set across different cohorts.
    # TODO: we're currently using the same isovar cache that we use for expressed
    # neoantigen prediction; so we pass in the same epitope lengths.
    # This is hacky and should be addressed.
    df_isovar = patient.cohort.load_single_patient_isovar(
        patient=patient,
        variants=variant_collection,
        epitope_lengths=[8, 9, 10, 11])
    expressed_variant_set = set()
    for _, row in df_isovar.iterrows():
        expressed_variant = Variant(contig=row["chr"],
                          start=row["pos"],
                          ref=row["ref"],
                          alt=row["alt"],
                          ensembl=variant_collection[0].ensembl)
        expressed_variant_set.add(expressed_variant)
    return expressed_variant_set

def variant_expressed_filter(filterable_variant, **kwargs):
    expressed_variants = expressed_variant_set(
        cohort=filterable_variant.patient.cohort,
        patient=filterable_variant.patient,
        variant_collection=filterable_variant.variant_collection)
    return filterable_variant.variant in expressed_variants

def effect_expressed_filter(filterable_effect, **kwargs):
    return variant_expressed_filter(filterable_effect, **kwargs)

def load_ensembl_coverage(cohort, coverage_path, min_tumor_depth, min_normal_depth=0,
                          pageant_dir_fn=None):
    """
    Load in Pageant CoverageDepth results with Ensembl loci.

    coverage_path is a path to Pageant CoverageDepth output directory, with
    one subdirectory per patient and a `cdf.csv` file inside each patient subdir.

    If min_normal_depth is 0, calculate tumor coverage. Otherwise, calculate
    join tumor/normal coverage.

    pageant_dir_fn is a function that takes in a Patient and produces a Pageant
    dir name.

    Last tested with Pageant CoverageDepth version 1ca9ed2.
    """
    # Function to grab the pageant file name using the Patient
    if pageant_dir_fn is None:
        pageant_dir_fn = lambda patient: patient.id

    columns_both = [
        "depth1", # Normal
        "depth2", # Tumor
        "onBP1",
        "onBP2",
        "numOnLoci",
        "fracBPOn1",
        "fracBPOn2",
        "fracLociOn",
        "offBP1",
        "offBP2",
        "numOffLoci",
        "fracBPOff1",
        "fracBPOff2",
        "fracLociOff",
    ]
    columns_single = [
        "depth",
        "onBP",
        "numOnLoci",
        "fracBPOn",
        "fracLociOn",
        "offBP",
        "numOffLoci",
        "fracBPOff",
        "fracLociOff"
    ]
    if min_normal_depth < 0:
        raise ValueError("min_normal_depth must be >= 0")
    use_tumor_only = (min_normal_depth == 0)
    columns = columns_single if use_tumor_only else columns_both
    ensembl_loci_dfs = []
    for patient in cohort:
        patient_ensembl_loci_df = pd.read_csv(
            path.join(coverage_path, pageant_dir_fn(patient), "cdf.csv"),
            names=columns,
            header=1)
        # pylint: disable=no-member
        # pylint gets confused by read_csv
        if use_tumor_only:
            depth_mask = (patient_ensembl_loci_df.depth == min_tumor_depth)
        else:
            depth_mask = (
                (patient_ensembl_loci_df.depth1 == min_normal_depth) &
                (patient_ensembl_loci_df.depth2 == min_tumor_depth))
        patient_ensembl_loci_df = patient_ensembl_loci_df[depth_mask]
        assert len(patient_ensembl_loci_df) == 1, (
            "Incorrect number of tumor={}, normal={} depth loci results: {} for patient {}".format(
                min_tumor_depth, min_normal_depth, len(patient_ensembl_loci_df), patient))
        patient_ensembl_loci_df["patient_id"] = patient.id
        ensembl_loci_dfs.append(patient_ensembl_loci_df)
    ensembl_loci_df = pd.concat(ensembl_loci_dfs)
    ensembl_loci_df["MB"] = ensembl_loci_df.numOnLoci / 1000000.0
    return ensembl_loci_df[["patient_id", "numOnLoci", "MB"]]



import os
import tempfile
from contextlib import contextmanager
from urllib.parse import urlparse

from google.cloud import storage


_GOOGLE_STORAGE_SCHEMA = "gs"

## Things that can go here ##
class GoogleStorageEnvironmentError(EnvironmentError): pass
class GoogleStorageValueError(ValueError): pass
class GoogleStorageIOError(IOError): pass
## End of things that can go wrong


class GoogleStorageIO:
    """
    This is a simple, light-weight wrapper aound Google Cloud's SDK
    that comes with pretty low-level REST-like operations. Given that
    we mostly depend on files sitting either on the NFS or the GS,
    I thought this could facilitate a few things and make it easier
    to switch between NFS and GCloud worlds.

    Note that Google's SDK tries its best to locate a configuration or
    authentication file around for it to access the data; but if
    you might need to use the following command once (if you haven't done so
    before) so that the auth part goes much smoother (basically hands-off):

    ```
        # Beta channel
        $ gcloud beta auth application-default login
        # Stable channel
        $ gcloud auth application-default login
    ```

    More information on the auth procedures:
        https://cloud.google.com/docs/authentication/
    """
    def __init__(self, *args, **kwargs):
        """
        You can pass authentication related configuration directly down to
        gcloud client, so if the default automated auth method doesn't work
        this will help you go with other options: CREDs, service accounts, etc.
        """
        try:
            self._client = storage.Client(*args, **kwargs)
        except Exception as e:
            err = ("Couldn't initialize Google Storage client, possibly due "
                   "to authentication issues. Please check your gcloud "
                   "auth setup: {}".format(e))
            raise GoogleStorageEnvironmentError(err)


    @property
    def client(self):
        """
        A standard Google Storage client. Exposing this carefully, because
        there might be helpful storage-based operations that are not
        implemented in this class; and this might come handy in such
        cases.
        """
        return self._client


    def list_files(self, gsuri, **kwargs):
        bucket_name, gs_rel_path = self.parse_uri(gsuri)
        bucket = self._client.get_bucket(bucket_name)
        return bucket.list_blobs(**kwargs)


    @client.setter
    def client(self, new_client):
        self._client = new_client


    def parse_uri(self, gsuri):
        gs_parts = urlparse(gsuri)
        assert gs_parts.scheme == _GOOGLE_STORAGE_SCHEMA
        bucket_name = gs_parts.netloc
        gs_rel_path = gs_parts.path
        # And apparently gcloud doesn't like relative URIs
        # that starts with a `/`, so let's get rid of that:
        gs_rel_path = gs_rel_path if gs_rel_path[0] != '/' else gs_rel_path[1::]
        return (bucket_name, gs_rel_path)


    def upload_file(self, localpath, gsuri):
        # And now request the handles for bucket and the file
        bucket_name, rel_path = self.parse_uri(gsuri)
        bucket = self.client.get_bucket(bucket_name)
        ublob = storage.Blob(rel_path, bucket)
        ublob.upload_from_filename(localpath)


    def download_to_path(self, gsuri, localpath, binary_mode=False, tmpdir=None):
        """
        This method is analogous to "gsutil cp gsuri localpath", but in a
        programatically accesible way. The only difference is that we
        have to make a guess about the encoding of the file to not upset
        downstream file operations. If you are downloading a VCF, then
        "False" is great. If this is a BAM file you are asking for, you
        should enable the "binary_mode" to make sure file doesn't get
        corrupted.

        gsuri: full GS-based URI, e.g. gs://cohorts/rocks.txt
        localpath: the path for the downloaded file, e.g. /mnt/cohorts/yep.txt
        binary_mode: (logical) if yes, the binary file operations will be
                     used; if not, standard ascii-based ones.
        """
        bucket_name, gs_rel_path = self.parse_uri(gsuri)
        # And now request the handles for bucket and the file
        bucket = self._client.get_bucket(bucket_name)
        # Just assignment, no downloading (yet)
        ablob = bucket.get_blob(gs_rel_path)
        if not ablob:
            raise GoogleStorageIOError(
                "No such file on Google Storage: '{}'".format(gs_rel_path))

        # A tmp file to serve intermediate phase
        # should be on same filesystem as localpath
        tmp_fid, tmp_file_path = tempfile.mkstemp(text=(not binary_mode),
                                                  dir=tmpdir)
        # set chunk_size to reasonable default
        # https://github.com/GoogleCloudPlatform/google-cloud-python/issues/2222
        ablob.chunk_size = 1<<30
        # Download starts in a sec....
        ablob.download_to_filename(client=self._client, filename=tmp_file_path)
        # ... end download ends. Let's move our finished file over.

        # You will see that below, instead of directly writing to a file
        # we are instead first using a different file and then move it to
        # its final location. We are doing this because we don't want
        # corrupted/incomplete data to be around as much as possible.
        return os.rename(tmp_file_path, localpath)

    def get_blob(self, gsuri):
        bucket_name, gs_rel_path = self.parse_uri(gsuri)
        bucket = self._client.get_bucket(bucket_name)
        ablob = bucket.get_blob(gs_rel_path)
        return ablob


class GoogleStorageFile(object):
    """
    This class is a context manager and a context decorator (future) that
    is meant to make working on GS files feel similar to local ones.
    Here is a one way via the `with` keyword:
        ```
        gsuri = 'gs://cohorts/biomarker.csv'
        with GoogleStorageFile(gsuri, 'r+') as fgs:
            contents = fgs.read()
            # since `fgs` is a local one, it is easy to
            # to work against it.
            # ...
            gfs.write(...)
        # And boom, all the tmp files are cleaned and our file is
        # uploaded back to the GS in its new form.
        ```
    """
    def __init__(self, gcio, gsuri, mode):
        self.gcio = gcio
        self.gsuri = gsuri
        self.is_read = 'r' in mode
        self.is_write = ('w' in mode) or ('r+' in mode)
        self.is_binary = 'b' in mode
        self.mode = mode
        self._localfile = None

    def __enter__(self):
        # Get a temp file and use it as a local file to immitate
        # direct GS access by downloading contents onto that file.
        tmp_fid, tmp_file_path = tempfile.mkstemp(text=(not self.is_binary))
        self.gcio.download_to_path(self.gsuri, tmp_file_path)
        self._localfile = open(tmp_file_path, self.mode)
        self._localfile_path = tmp_file_path
        return self._localfile

    def __exit__(self, *args):
        # We are done with the open file, so let's close it
        self._localfile.close()
        # If write mode is on, then upload the altered one to GS
        if self.is_write:
            bucket_name, rel_path = self.gcio.parse_uri(self.gsuri)
            bucket = self.gcio.client.get_bucket(bucket_name)
            ublob = storage.Blob(rel_path, bucket)
            ublob.upload_from_filename(self._localfile_path)
        # And because of the type of the temp file we created,
        # we are responsible for deleting it when we are finshed:
        os.remove(self._localfile_path)



# Copyright (c) 2017. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from . import generated_data_path
from cohorts import Patient, Cohort
from varcode import Variant, VariantCollection
from varcode.effects import Substitution, FrameShift, IntronicSpliceSite
import pandas as pd
from nose.tools import eq_, ok_

def test_splice_filtering_substitution():
    """
    Make sure that ExonicSpliceSite mutations with Substitution alternates are kept even when we filter to
    only Substitution effects.
    """
    cohort = None
    try:
        variant = Variant(contig=10, start=124340409, ref="C", alt="A", ensembl=75)
        patient = Patient(id="patient", os=3, pfs=2, deceased=False, progressed=False, variants=VariantCollection([variant]))
        cohort_cache_path = generated_data_path("cache")
        cohort = Cohort(
            patients=[patient],
            cache_dir=cohort_cache_path)

        def missense_snv_filter(filterable_effect):
            return (type(filterable_effect.effect) == Substitution and
                    filterable_effect.variant.is_snv)

        effects = cohort.load_effects(filter_fn=missense_snv_filter)[patient.id]
        eq_(len(effects), 1)
        all_effects = cohort.load_effects(filter_fn=missense_snv_filter,
                                          all_effects=True)[patient.id]
        eq_(len(all_effects), 7)
    finally:
        if cohort is not None:
            cohort.clear_caches()

def test_splice_filtering_frameshift():
    """
    Make sure that ExonicSpliceSite mutations with FrameShift alternates are kept even when we filter to
    only FrameShift effects.
    """
    cohort = None
    try:
        variant = Variant(contig=8, start=145617535, ref="GGGGGTGCAAGGTGA", alt="", ensembl=75)
        patient = Patient(id="patient", os=3, pfs=2, deceased=False, progressed=False, variants=VariantCollection([variant]))
        cohort_cache_path = generated_data_path("cache")
        cohort = Cohort(
            patients=[patient],
            cache_dir=cohort_cache_path)

        def frameshift_filter(filterable_effect):
            return (type(filterable_effect.effect) == FrameShift)

        effects = cohort.load_effects(filter_fn=frameshift_filter)[patient.id]
        eq_(len(effects), 1)
        all_effects = cohort.load_effects(filter_fn=frameshift_filter,
                                          all_effects=True)[patient.id]
        eq_(len(all_effects), 1)
    finally:
        if cohort is not None:
            cohort.clear_caches()

def test_effects_priority_caching():
    """
    Make sure that effects are cached such that they are not filtered
    prematurely. See https://github.com/hammerlab/cohorts/issues/252.
    """
    cohort = None
    try:
        # This variant has IntronicSpliceSite, Subsitution effects, and more.
        variant = Variant(contig=3, start=20212211, ref="C", alt="T", ensembl=75)
        patient = Patient(id="patient", os=3, pfs=2, deceased=False, progressed=False, variants=VariantCollection([variant]))
        cohort_cache_path = generated_data_path("cache")
        cohort = Cohort(
            patients=[patient],
            cache_dir=cohort_cache_path)

        # All of the effects.
        cohort.clear_caches()
        for i in range(2):
            effects = cohort.load_effects(all_effects=True)[patient.id]
            eq_(len(effects), 15)

        # Top priority effect.
        cohort.clear_caches()
        for i in range(2):
            effects = cohort.load_effects()[patient.id]
            eq_(len(effects), 1)
            eq_(type(effects[0]), IntronicSpliceSite)

        def missense_snv_filter(filterable_effect):
            return (type(filterable_effect.effect) == Substitution and
                    filterable_effect.variant.is_snv)

        # All missense SNV effects, from the large cache.
        cohort.clear_caches()
        for i in range(2):
            effects = cohort.load_effects(all_effects=True, filter_fn=missense_snv_filter)[patient.id]
            eq_(len(effects), 6)

        # Top missense SNV effect, from the large cache.
        cohort.clear_caches()
        for i in range(2):
            effects = cohort.load_effects(filter_fn=missense_snv_filter)[patient.id]
            eq_(len(effects), 1)
            eq_(type(effects[0]), Substitution)

        # Top missense SNV effects, from the small nonsynonymous cache.
        cohort.clear_caches()
        for i in range(2):
            effects = cohort.load_effects(only_nonsynonymous=True, filter_fn=missense_snv_filter)[patient.id]
            eq_(len(effects), 1)
            eq_(type(effects[0]), Substitution)

        # All nonsynonymous effects, from the small nonsynonymous cache.
        cohort.clear_caches()
        for i in range(2):
            effects = cohort.load_effects(all_effects=True, only_nonsynonymous=True)[patient.id]
            eq_(len(effects), 6)
    finally:
        if cohort is not None:
            cohort.clear_caches()



# Copyright (c) 2017. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings

class DataFrameLoader(object):
    """
    Wraps a `DataFrame` with some information on how to join it.

    Parameters
    __________
    name : str
        The name of the dataframe, to easily reference it.
    load_dataframe : function
        A function that returns the `DataFrame` object.
    join_on_right : str
        The column of the `DataFrame` to join on (i.e. the patient
        ID column name).
    join_on_left: str
        The corresponding column in the `cohorts.Patient.additional_data`
	to join on, if not the `id` (which is the default).
    """
    def __init__(self,
                 name,
                 load_dataframe,
                 join_on=None,
                 join_on_right="patient_id",
                 join_on_left="patient_id"):
        self.name = name
        self.load_dataframe = load_dataframe
        if join_on is not None:
            warnings.warn("`join_on` parameter is deprecated. Please use `join_on_right` instead.", DeprecationWarning)
            self.join_on_right = join_on
        else:
            self.join_on_right = join_on_right
        self.join_on_left = join_on_left
